from __future__ import annotations

from typing import Literal

from ain.ain import Ain

from .chat import Chat


class Ainft:
    def __init__(
        self,
        *,
        private_key: str,
        api_url: str,
        blockchain_url: str,
        chain_id: Literal[0, 1],
    ):
        self._base_url = api_url
        self._blockchain_url = blockchain_url
        self._chain_id = chain_id

        self._ain = Ain(self._blockchain_url, self._chain_id)
        self._set_default_account(private_key)

        self.chat = Chat(self._ain)

    def _set_default_account(self, private_key: str):
        self._ain.wallet.clear()
        self._ain.wallet.addAndSetDefaultAccount(private_key)



from pydantic import BaseModel
from typing import Any, List, Literal, Optional, Union

SetMultiOperationType = Literal["SET"]
GetMultiOperationType = Literal["GET"]
SetOperationType = Literal[
    "SET_VALUE", "INC_VALUE", "DEC_VALUE", "SET_RULE", "SET_OWNER", "SET_FUNCTION"
]
GetOperationType = Literal["GET_VALUE", "GET_RULE", "GET_OWNER", "GET_FUNCTION"]


class SetOperation(BaseModel):
    type: SetOperationType
    ref: str
    value: Optional[Any] = None


class SetMultiOperation(BaseModel):
    type: SetMultiOperationType
    op_list: List[SetOperation]


class Thread(BaseModel):
    id: str
    metadata: Optional[dict] = None
    created_at: int


class Message(BaseModel):
    id: str
    """
    The ID of the message, 
    must be a 20-character alphanumeric starting with 'msg_' or a digit string.
    """
    thread_id: str
    """The ID of the thread."""
    role: Literal["user", "assistant"]
    """The role of the entity creating the message."""
    content: str
    """
    The content of the message. 
    If the content exceeds 200 characters, it will be automatically trimmed.
    """
    assistant_id: Optional[str] = None
    """If applicable, the ID of the assistant that authored this message."""
    run_id: Optional[str] = None
    """If applicable, the ID of the run associated with the authoring of this message."""
    metadata: Optional[dict] = None
    """
    The metadata can contain up to 16 key-value pairs,
    with the keys limited to 64 characters and the values to 512 characters.
    """
    created_at: int
    """The UNIX timestamp in seconds, used as a key."""


class TransactionInput(BaseModel):
    operation: Union[SetOperation, SetMultiOperation]
    timestamp: int
    nonce: Optional[int] = None
    address: Optional[str] = None
    gas_price: Optional[int] = None


class TransactionResult(BaseModel):
    tx_hash: str
    result: dict


class ThreadTransactionResult(TransactionResult):
    thread: Thread


class MessageTransactionResult(TransactionResult):
    messages: List[Message]



from __future__ import annotations

from datetime import datetime
import re
import uuid
from typing import Any


from ain.ain import Database, Wallet


def get_app_id(object_id: str) -> str:
    return f"ainft721_{object_id.lower()}"


def get_user_address(wallet: Wallet) -> str | None:
    return getattr(wallet.defaultAccount, "address", None)


def is_tx_success(tx_result: Any) -> bool:
    result = tx_result.get("result", {})
    if result.get("code") is not None and result["code"] != 0:
        return False
    if "result_list" in result:
        results = result["result_list"].values()
        return all(r.get("code") == 0 for r in results)
    return True


def join_paths(paths: list) -> str:
    return "/".join(paths)


def now() -> float:
    return datetime.now().timestamp()


def validate_user_address(wallet: Wallet) -> str:
    user_addr = get_user_address(wallet)
    if user_addr is None:
        raise ValueError(f"The default account not found.")
    return user_addr


async def validate_app(app_id: str, db: Database):
    app_path = join_paths(["apps", app_id])
    app = await db.ref(app_path).getValue()
    if app is None:
        raise ValueError(f"The app {app_id} does not exist.")


async def validate_token(app_id: str, token_id: str, db: Database):
    token_path = join_paths(["apps", app_id, "tokens", token_id])
    token = await db.ref(token_path).getValue()
    if token is None:
        raise ValueError(f"The token {token_id} does not exist.")


async def validate_thread(
    app_id: str, token_id: str, address: str, thread_id: str, db: Database
):
    thread_path = join_paths(
        [
            "apps",
            app_id,
            "tokens",
            token_id,
            "ai",
            "ainize_openai",
            "history",
            address,
            "threads",
            thread_id,
        ]
    )
    thread = await db.ref(thread_path).getValue()
    if thread is None:
        raise ValueError(f"The thread {thread_id} does not exist.")


def validate_thread_id(thread_id: str):
    if not is_valid_thread_id(thread_id):
        raise ValueError(f"Invalid thread ID.")


def is_valid_thread_id(thread_id: str) -> bool:
    try:
        return str(uuid.UUID(thread_id, version=4)) == thread_id
    except ValueError:
        pass
    pattern = r"thread_[A-Za-z0-9]{20}$"
    if re.match(pattern, thread_id):
        return True
    return False


def validate_message_id(message_id: str):
    if not is_valid_message_id(message_id):
        raise ValueError(f"Invalid message ID.")


def is_valid_message_id(message_id: str) -> bool:
    if message_id.isdigit():
        return True
    pattern = r"msg_[A-Za-z0-9]{20}$"
    if re.match(pattern, message_id):
        return True
    return False



from __future__ import annotations

import json
from typing import List

from ain.ain import Ain

from ..types import (
    SetOperation,
    SetMultiOperation,
    TransactionInput,
    Message,
    MessageTransactionResult,
)
from ..utils import *


class Messages:
    _max_content_length = 200

    def __init__(self, ain: Ain):
        self._ain = ain

    async def store(
        self,
        *,
        messages: List[Message],
        object_id: str,
        token_id: str,
    ) -> MessageTransactionResult:
        """
        Store a list of messages.

        Args:
            messages: The list of messages.

            object_id: The ID of the AINFT object.

            token_id: The ID of the AINFT token.
        """
        app_id = get_app_id(object_id)
        user_addr = validate_user_address(self._ain.wallet)

        await self._validate(app_id, token_id, user_addr, messages)

        return await self._send_tx_for_store_message(
            **dict(
                messages=messages,
                app_id=app_id,
                token_id=token_id,
                address=user_addr,
            )
        )

    async def _validate(
        self, app_id: str, token_id: str, address: str, messages: List[Message]
    ):
        await validate_app(app_id, self._ain.db)
        await validate_token(app_id, token_id, self._ain.db)
        for message in messages:
            await validate_thread(
                app_id, token_id, address, message.thread_id, self._ain.db
            )
            validate_message_id(message.id)

    async def _send_tx_for_store_message(self, **kwargs) -> MessageTransactionResult:
        timestamp = int(now())
        tx_body = self._build_tx_body_for_store_message(timestamp=timestamp, **kwargs)
        tx_result = await self._ain.sendTransaction(tx_body)

        if not is_tx_success(tx_result):
            raise RuntimeError(f"Failed to send transaction: {json.dumps(tx_result)}")

        return self._format_tx_result(tx_result=tx_result, **kwargs)

    def _build_tx_body_for_store_message(
        self,
        app_id: str,
        token_id: str,
        messages: List[Message],
        address: str,
        timestamp: int,
    ) -> TransactionInput:
        op_list = []
        for msg in messages:
            msg_key = str(msg.created_at)
            msg_path = join_paths(
                [
                    "apps",
                    app_id,
                    "tokens",
                    token_id,
                    "ai",
                    "ainize_openai",
                    "history",
                    address,
                    "threads",
                    msg.thread_id,
                    "messages",
                    msg_key,
                ]
            )
            trimmed_content = msg.content[: self._max_content_length]
            message = {
                "id": msg.id,
                "role": msg.role,
                "content": trimmed_content,
                **({"metadata": msg.metadata} if msg.metadata else {}),
            }
            op = SetOperation(type="SET_VALUE", ref=msg_path, value=message)
            op_list.append(op)

        multi_op = SetMultiOperation(type="SET", op_list=op_list)
        return TransactionInput(
            operation=multi_op,
            timestamp=timestamp,
            nonce=-1,
            address=address,
            gas_price=500,
        )

    def _format_tx_result(
        self, tx_result: dict, messages: List[Message], **kwargs
    ) -> MessageTransactionResult:
        return MessageTransactionResult(messages=messages, **tx_result)



from __future__ import annotations

from functools import cached_property

from ain.ain import Ain

from .threads import Threads
from .messages import Messages


class Chat:
    def __init__(self, ain: Ain) -> None:
        self._ain = ain

    @cached_property
    def threads(self) -> Threads:
        return Threads(self._ain)

    @cached_property
    def messages(self) -> Messages:
        return Messages(self._ain)



from __future__ import annotations

from .client import Ainft
from .types import (
    Thread,
    Message,
    TransactionResult,
    ThreadTransactionResult,
    MessageTransactionResult,
)

__all__ = [
    "Ainft",
    "Thread",
    "Message",
    "TransactionResult",
    "ThreadTransactionResult",
    "MessageTransactionResult",
]



from __future__ import annotations

import json
from typing import Optional

from ain.ain import Ain

from ..types import SetOperation, TransactionInput, Thread, ThreadTransactionResult
from ..utils import *


class Threads:
    def __init__(self, ain: Ain):
        self._ain = ain

    async def store(
        self,
        *,
        thread_id: str,
        object_id: str,
        token_id: str,
        metadata: Optional[dict] = None,
    ) -> ThreadTransactionResult:
        """
        Store a thread.

        Args:
            thread_id: The ID of the thread,
                must be a 20-character alphanumeric starting with 'thread_' or a uuid4.

            object_id: The ID of the AINFT object.

            token_id: The ID of the AINFT token.

            metadata: The metadata can contain up to 16 key-value pairs,
                with the keys limited to 64 characters and the values to 512 characters.
        """
        app_id = get_app_id(object_id)
        user_addr = validate_user_address(self._ain.wallet)

        await self._validate(app_id, token_id, thread_id)

        return await self._send_tx_for_store_thread(
            **dict(
                thread_id=thread_id,
                app_id=app_id,
                token_id=token_id,
                address=user_addr,
                metadata=metadata,
            )
        )

    async def _validate(self, app_id: str, token_id: str, thread_id: str):
        await validate_app(app_id, self._ain.db)
        await validate_token(app_id, token_id, self._ain.db)
        validate_thread_id(thread_id)

    async def _send_tx_for_store_thread(self, **kwargs) -> ThreadTransactionResult:
        timestamp = int(now())
        tx_body = self._build_tx_body_for_store_thread(timestamp=timestamp, **kwargs)
        tx_result = await self._ain.sendTransaction(tx_body)

        if not is_tx_success(tx_result):
            raise RuntimeError(f"Failed to send transaction: {json.dumps(tx_result)}")

        return self._format_tx_result(
            tx_result=tx_result, timestamp=timestamp, **kwargs
        )

    def _build_tx_body_for_store_thread(
        self,
        app_id: str,
        token_id: str,
        thread_id: str,
        address: str,
        timestamp: int,
        metadata: Optional[dict],
    ) -> TransactionInput:
        thread_path = join_paths(
            [
                "apps",
                app_id,
                "tokens",
                token_id,
                "ai",
                "ainize_openai",
                "history",
                address,
                "threads",
                thread_id,
            ]
        )
        thread = {
            "messages": True,
            **({"metadata": metadata} if metadata else {}),
        }
        op = SetOperation(type="SET_VALUE", ref=thread_path, value=thread)
        return TransactionInput(
            operation=op,
            timestamp=timestamp,
            nonce=-1,
            address=address,
            gas_price=500,
        )

    def _format_tx_result(
        self,
        tx_result: dict,
        timestamp: int,
        thread_id: str,
        **kwargs,
    ) -> ThreadTransactionResult:
        metadata = kwargs.get("metadata", {})
        thread = Thread(id=thread_id, metadata=metadata, created_at=timestamp)
        return ThreadTransactionResult(thread=thread, **tx_result)
